# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import importlib
import json
import os
import random

import numpy as np
import pytest
import torch

eval_model = importlib.import_module("compressai.utils.video.eval_model.__main__")
update_model = importlib.import_module("compressai.utils.update_model.__main__")

# Example: GENERATE_EXPECTED=1 pytest -sx tests/test_eval_model.py
GENERATE_EXPECTED = os.getenv("GENERATE_EXPECTED")


def set_rng_seed(seed):
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)


def test_eval_model_video():
    with pytest.raises(SystemExit):
        eval_model.main(["--help"])

    with pytest.raises(SystemExit):
        eval_model.main([])

    with pytest.raises(SystemExit):
        eval_model.main(["pretrained"])

    with pytest.raises(SystemExit):
        eval_model.main(
            [
                "pretrained",
                ".",
                "-a",
                "ssf2020",
                "-m",
                "mse",
                "-q",
                "1",
            ]
        )


# mse and entropy_estimation tested for now
@pytest.mark.parametrize("model", ("ssf2020",))
@pytest.mark.parametrize("quality", ("1", "4", "8"))
@pytest.mark.parametrize("metric", ("mse",))
@pytest.mark.parametrize("entropy_estimation", (True, False))
def test_eval_model_pretrained(model, quality, metric, entropy_estimation, tmpdir):
    here = os.path.dirname(__file__)
    dirpath = os.path.join(here, "assets/dataset/video")

    cmd = [
        "pretrained",
        dirpath,
        str(tmpdir),
        "-a",
        model,
        "-m",
        metric,
        "-q",
        quality,
        "-d",
        str(tmpdir),
        "-o",
        f"{model}-{metric}-{quality}",
    ]
    if entropy_estimation:
        cmd += ["--entropy-estimation"]
    eval_model.main(cmd)

    with open(f"{tmpdir}/{model}-{metric}-{quality}.json") as f:
        output = json.load(f)

    expected = os.path.join(
        here,
        "expected",
        f"eval_{int(entropy_estimation)}_{model}_{metric}_{quality}.json",
    )

    if not os.path.isfile(expected):
        if not GENERATE_EXPECTED:
            raise RuntimeError(f"Missing expected file {expected}")
        with open(expected, "w") as f:
            json.dump(output, f)

    with open(expected, "r") as f:
        expected = json.loads(f.read())

    for key in ("name", "description"):
        assert expected[key] == output[key]

    for key in (
        "psnr-y",
        "psnr-u",
        "psnr-v",
        "psnr-yuv",
        "psnr-rgb",
        "ms-ssim-rgb",
        "bitrate",
    ):
        if key not in expected["results"]:
            continue
        assert np.allclose(
            expected["results"][key], output["results"][key], rtol=1e-4, atol=1e-4
        )


# @pytest.mark.parametrize("model_name", ("ssf2020",))
# def test_eval_model_ckpt(tmp_path, model_name):
#     here = os.path.dirname(__file__)
#     parent = os.path.dirname(here)

#     # fake training
#     datapath = os.path.join(here, "assets/fakedata/imagefolder")
#     spec = importlib.util.spec_from_file_location(
#         "examples.train", os.path.join(parent, "examples/train_video.py")
#     )
#     module = importlib.util.module_from_spec(spec)
#     spec.loader.exec_module(module)

#     argv = [
#         "-d",
#         datapath,
#         "-e",
#         "1",
#         "--batch-size",
#         "1",
#         "--patch-size",
#         "48",
#         "64",
#         "--seed",
#         "0",
#         "--save",
#     ]

#     os.chdir(tmp_path)
#     module.main(argv)

#     checkpoint = "checkpoint_best_loss.pth.tar"
#     assert os.path.isfile(checkpoint)

#     # update model
#     cmd = ["-a", model_name, "-n", "factorized", checkpoint]
#     update_model.main(cmd)

#     # ckpt evaluation
#     dirpath = os.path.join(here, "assets/dataset/image")
#     checkpoint = next(f for f in os.listdir(tmp_path) if f.startswith("factorized-"))
#     cmd = [
#         "checkpoint",
#         dirpath,
#         "-a",
#         "bmshj2018-factorized",
#         "-p",
#         checkpoint,
#     ]
#     eval_model.main(cmd)



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from torch.hub import load_state_dict_from_url

from compressai.models import (
    JointAutoregressiveHierarchicalPriorsVbr,
    MeanScaleHyperpriorVbr,
    ScaleHyperpriorVbr,
)

from .pretrained import load_pretrained

__all__ = [
    "bmshj2018_hyperprior_vbr",
    "mbt2018_mean_vbr",
    "mbt2018_vbr",
]

model_architectures = {
    "bmshj2018-hyperprior-vbr": ScaleHyperpriorVbr,
    "mbt2018-mean-vbr": MeanScaleHyperpriorVbr,
    "mbt2018-vbr": JointAutoregressiveHierarchicalPriorsVbr,
}

root_url = "https://compressai.s3.amazonaws.com/models/v1"
model_urls = {
    "bmshj2018-hyperprior-vbr": {
        "mse": f"{root_url}/bmshj2018-hyperprior-mse-vbr-cddd26be.pth.tar",
        # "ms-ssim": f"{root_url}/bmshj2018-hyperprior-ms-ssim-vbr-HASH.pth.tar",
    },
    "mbt2018-mean-vbr": {
        "mse": f"{root_url}/mbt2018-mean-mse-vbr-45d095e9.pth.tar",
        # "ms-ssim": f"{root_url}/mbt2018-mean-ms-ssim-vbr-HASH.pth.tar",
    },
    "mbt2018-vbr": {
        "mse": f"{root_url}/mbt2018-mse-vbr-f12581a1.pth.tar",
        # "ms-ssim": f"{root_url}/mbt2018-ms-ssim-vbr-HASH.pth.tar",
    },
}

cfgs = {
    "bmshj2018-hyperprior-vbr": (192, 320),
    "mbt2018-mean-vbr": (192, 320),
    "mbt2018-vbr": (192, 320),
}


def _load_model(architecture, metric, pretrained=False, progress=True, **kwargs):
    if architecture not in model_architectures:
        raise ValueError(f'Invalid architecture name "{architecture}"')

    if pretrained:
        if architecture not in model_urls or metric not in model_urls[architecture]:
            raise RuntimeError("Pre-trained model not yet available")

        url = model_urls[architecture][metric]
        state_dict = load_state_dict_from_url(url, progress=progress)
        state_dict = load_pretrained(state_dict)
        if architecture in ["bmshj2018-hyperprior-vbr", "mbt2018-mean-vbr"]:
            model = model_architectures[architecture].from_state_dict(
                state_dict, vr_entbttlnck=True
            )
        else:
            model = model_architectures[architecture].from_state_dict(state_dict)
        return model

    model = model_architectures[architecture](*cfgs[architecture], **kwargs)
    return model


def bmshj2018_hyperprior_vbr(
    quality=0, metric="mse", pretrained=False, progress=True, **kwargs
):
    r"""Variable bitrate (vbr) version of bmshj2018-hyperprior (see compressai/models/google.py) with variable bitrate components detailed in:
    Fatih Kamisli, Fabien Racape and Hyomin Choi
    `"Variable-Rate Learned Image Compression with Multi-Objective Optimization and Quantization-Reconstruction Offsets`"
    <https://arxiv.org/abs/2402.18930>`_, Data Compression Conference (DCC), 2024.

    Args:
        metric (str): Optimized metric, choose from ('mse', no 'ms-ssim' yet)
        pretrained (bool): If True, returns a pre-trained model
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    # keep quality here for now for consistency with image models.
    # could be used to indicate which lambda or quality was used to train the base model.
    del quality
    if metric not in ("mse"):  # ("mse", "ms-ssim"): # we have only mse model
        raise ValueError(f'Invalid metric "{metric}"')

    return _load_model(
        "bmshj2018-hyperprior-vbr", metric, pretrained, progress, **kwargs
    )


def mbt2018_mean_vbr(
    quality=0, metric="mse", pretrained=False, progress=True, **kwargs
):
    r"""Variable bitrate (vbr) version of bmshj2018 (see compressai/models/google.py) with variable bitrate components detailed in:
    Fatih Kamisli, Fabien Racape and Hyomin Choi
    `"Variable-Rate Learned Image Compression with Multi-Objective Optimization and Quantization-Reconstruction Offsets`"
    <https://arxiv.org/abs/2402.18930>`_, Data Compression Conference (DCC), 2024.

    Args:
        metric (str): Optimized metric, choose from ('mse', no 'ms-ssim' yet)
        pretrained (bool): If True, returns a pre-trained model
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    # keep quality here for now for consistency with image models.
    # could be used to indicate which lambda or quality was used to train the base model.
    del quality
    if metric not in ("mse"):  # ("mse", "ms-ssim"): # we have only mse model
        raise ValueError(f'Invalid metric "{metric}"')

    return _load_model("mbt2018-mean-vbr", metric, pretrained, progress, **kwargs)


def mbt2018_vbr(quality=0, metric="mse", pretrained=False, progress=True, **kwargs):
    r"""Variable bitrate (vbr) version of mbt2018 (see compressai/models/google.py) with variable bitrate components detailed in:
    Fatih Kamisli, Fabien Racape and Hyomin Choi
    `"Variable-Rate Learned Image Compression with Multi-Objective Optimization and Quantization-Reconstruction Offsets`"
    <https://arxiv.org/abs/2402.18930>`_, Data Compression Conference (DCC), 2024.

    Args:
        metric (str): Optimized metric, choose from ('mse', no 'ms-ssim' yet)
        pretrained (bool): If True, returns a pre-trained model
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    # keep quality here for now for consistency with image models.
    # could be used to indicate which lambda or quality was used to train the base model.
    del quality
    if metric not in ("mse"):  # ("mse", "ms-ssim"): # we have only mse model
        raise ValueError(f'Invalid metric "{metric}"')

    return _load_model("mbt2018-vbr", metric, pretrained, progress, **kwargs)



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import pickle
import re
import sys

# expect pickle file format:
# {
#   'quantizers': { 'weight0': 10, 'weight1': 13 ...},
# }
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="extract_quantizers",
        usage="use output of debug_model to identify layers index.\nusage: debug_model model.sadl | extract_quantizers --input model_info.pkl > quantizers.txt",
    )
    parser.add_argument(
        "--input",
        action="store",
        nargs="?",
        type=str,
        help="name of the pkl file containing the dict",
    )
    parser.add_argument("--verbose", action="count")

args = parser.parse_args()

with open(args.input, "rb") as f:
    infos = pickle.load(f)

if args.verbose:
    print(infos, file=sys.stderr)

started = False
d = {}
firstLayer = None
conv = {}
for line in sys.stdin:
    if "Exit" == line.rstrip() or "end model inference" in line:
        break
    if "start model inference" in line:
        started = True
    if started:
        m = re.match(r"\[INFO\]\s+(\d+)\s+Const\s+\((.+)\).*", line)
        if m is not None:
            d[m[2]] = int(m[1])

        m = re.match(r"\[INFO\]\s+(\d+)\s+Conv2DTranspose\s+\((.+)\).*", line)
        if m is not None:
            if firstLayer is None:
                firstLayer = int(m[1])
            else:
                conv[int(m[1])] = 0

if args.verbose:
    print(
        "[INFO] first layer: {}, layers index: {}".format(firstLayer, d),
        file=sys.stderr,
    )

assert "quantizers" in infos, "No quantizers key in {}".format(args.input)

# get quantizers
print("0 0 ", end="")  # input is already in integer, do not put any quantizer
for k, v in infos["quantizers"].items():
    name = k
    assert name in d, "{} not in extracted names {}".format(name, d)
    print("{} {} ".format(d[name], int(v)), end="")
    if int(d[name]) == firstLayer - 1:  # const layer corresponding to the first deconv
        print("{} {} ".format(firstLayer, -int(v)), end="")
for k, v in conv.items():
    print("{} {} ".format(k, v), end="")
print("")



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import abc
import argparse
import platform
import subprocess
import sys

from pathlib import Path
from typing import Any, List

from compressai.datasets.rawvideo import RawVideoSequence, get_raw_video_file_info


def run_command(cmd, ignore_returncodes=None):
    cmd = [str(c) for c in cmd]
    try:
        rv = subprocess.check_output(cmd)
        return rv.decode("ascii")
    except subprocess.CalledProcessError as err:
        if ignore_returncodes is not None and err.returncode in ignore_returncodes:
            return err.output
        print(err.output.decode("utf-8"))
        sys.exit(1)


class Codec(abc.ABC):
    name: str
    description: str
    help = ""

    @property
    @abc.abstractmethod
    def name(self):
        raise NotImplementedError()

    @property
    def description(self):
        return self._description

    @abc.abstractmethod
    def add_parser_args(self, parser: argparse.ArgumentParser) -> None:
        pass

    def set_args(self, args):
        return args

    @abc.abstractmethod
    def get_bin_path(self, filepath: Path, inputdir: Path, **args: Any) -> Path:
        raise NotImplementedError

    @abc.abstractmethod
    def get_encode_cmd(
        self, filepath: Path, qp: int, binpath: Path, **args: Any
    ) -> List[Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_decode_cmd(self, filepath: Path, **args: Any) -> List[Any]:
        raise NotImplementedError


def get_ffmpeg_version():
    rv = run_command(["ffmpeg", "-version"])
    return rv.split()[2]


class x264(Codec):
    preset = ""
    tune = ""

    @property
    def name(self):
        return "x264"

    @property
    def description(self):
        return f"{self.name} {self.preset}, {self.tune}, ffmpeg version {get_ffmpeg_version()}"

    def name_config(self):
        return f"{self.name}-{self.preset}-tune-{self.tune}"

    def add_parser_args(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("-p", "--preset", default="medium", help="preset")
        parser.add_argument(
            "--tune",
            default="psnr-rgb",
            help="tune encoder for psnr or ssim (default: %(default)s)",
        )

    def set_args(self, args):
        args = super().set_args(args)
        self.preset = args.preset
        self.tune = args.tune
        return args

    def get_bin_path(self, filepath: Path, qp, binpath: str, inputdir: Path) -> Path:
        bin_subdir = Path(binpath) / Path(filepath).parent.relative_to(inputdir)
        bin_subdir.mkdir(parents=True, exist_ok=True)
        return bin_subdir / (
            f"{filepath.stem}_{self.name}_{self.preset}_tune-{self.tune}_qp{qp}.mp4"
        )

    def get_encode_cmd(self, filepath: Path, qp: int, binpath: Path) -> List[Any]:
        info = get_raw_video_file_info(filepath.stem)
        cmd = [
            "ffmpeg",
            "-y",
            "-s:v",
            f"{info['width']}x{info['height']}",
            "-i",
            filepath,
            "-c:v",
            "h264",
            "-crf",
            qp,
            "-preset",
            self.preset,
            "-bf",
            0,
            "-tune",
            self.tune,
            "-pix_fmt",
            "yuv420p",
            "-threads",
            "4",
            binpath,
        ]
        return cmd

    def get_decode_cmd(
        self, binpath: Path, decpath: Path, input_filepath: Path
    ) -> List[Any]:
        del input_filepath  # unused here
        cmd = [
            "ffmpeg",
            "-y",
            "-i",
            binpath,
            "-pix_fmt",
            "yuv420p",
            decpath,
        ]
        return cmd


class x265(x264):
    @property
    def name(self):
        return "x265"

    def get_encode_cmd(self, filepath: Path, qp: int, binpath: Path) -> List[Any]:
        info = get_raw_video_file_info(filepath.stem)
        cmd = [
            "ffmpeg",
            "-s:v",
            f"{info['width']}x{info['height']}",
            "-i",
            filepath,
            "-c:v",
            "hevc",
            "-crf",
            qp,
            "-preset",
            self.preset,
            "-x265-params",
            "bframes=0",
            "-tune",
            self.tune,
            "-pix_fmt",
            "yuv420p",
            "-threads",
            "4",
            binpath,
        ]
        return cmd


class VTM(Codec):
    """VTM: VVC reference software"""

    binext = "bin"
    config = ""

    @property
    def name(self):
        return "VTM"

    @property
    def description(self):
        return f"VTM reference software, version {self.get_version(self.encoder_path)}"

    def name_config(self):
        return f"{self.name}-v{self.get_version(self.encoder_path)}-{self.config}"

    def get_version(selfm, encoder_path):
        rv = run_command([encoder_path, "--help"], ignore_returncodes=[1])
        version = rv.split(b"\n")[1].split()[4].decode().strip("[]")
        return version

    def get_encoder_path(self, build_dir):
        system = platform.system()
        try:
            elfnames = {"Darwin": "EncoderApp", "Linux": "EncoderAppStatic"}
            return Path(build_dir) / elfnames[system]
        except KeyError as err:
            raise RuntimeError(f'Unsupported platform "{system}"') from err

    def get_decoder_path(self, build_dir):
        system = platform.system()
        try:
            elfnames = {"Darwin": "DecoderApp", "Linux": "DecoderAppStatic"}
            return Path(build_dir) / elfnames[system]
        except KeyError as err:
            raise RuntimeError(f'Unsupported platform "{system}"') from err

    @classmethod
    def add_parser_args(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-b",
            "--build-dir",
            type=str,
            required=True,
            help="VTM build dir",
        )
        parser.add_argument(
            "-c",
            "--config",
            type=str,
            required=True,
            help="VTM config file",
        )
        parser.add_argument(
            "--rgb", action="store_true", help="Use RGB color space (over YCbCr)"
        )

    def set_args(self, args):
        args = super().set_args(args)
        self.encoder_path = self.get_encoder_path(args.build_dir)
        self.decoder_path = self.get_decoder_path(args.build_dir)
        self.config_path = args.config
        self.config = Path(self.config_path).stem.split("_")[1]
        self.version = self.get_version(self.encoder_path)
        self.rgb = args.rgb
        return args

    def get_encode_cmd(self, filepath: Path, qp, binpath) -> List[Any]:
        info = get_raw_video_file_info(filepath.stem)
        num_frames = len(RawVideoSequence.from_file(str(filepath)))
        cmd = [
            self.encoder_path,
            "-i",
            filepath,
            "-c",
            self.config_path,
            "-q",
            qp,
            "-o",
            "/dev/null",
            "-b",
            binpath,
            "-wdt",
            info["width"],
            "-hgt",
            info["height"],
            "-fr",
            info["framerate"],
            "-f",
            num_frames,
            f'--InputBitDepth={info["bitdepth"]}',
            f'--OutputBitDepth={info["bitdepth"]}',
            # "--ConformanceWindowMode=1",
        ]

        if self.rgb:
            cmd += [
                "--InputColourSpaceConvert=RGBtoGBR",
                "--SNRInternalColourSpace=1",
                "--OutputInternalColourSpace=0",
            ]
        return cmd

    def get_bin_path(self, filepath: Path, qp, binpath: str, inputdir: Path) -> Path:
        bin_subdir = Path(binpath) / Path(filepath).parent.relative_to(inputdir)
        bin_subdir.mkdir(parents=True, exist_ok=True)
        return bin_subdir / (
            f"{filepath.stem}_{self.name}_{self.config}_qp{qp}.{self.binext}"
        )

    def get_decode_cmd(
        self, binpath: Path, decpath: Path, input_filepath: Path
    ) -> List[Any]:
        output_bitdepth = get_raw_video_file_info(input_filepath.stem)["bitdepth"]
        cmd = [self.decoder_path, "-b", binpath, "-o", decpath, "-d", output_bitdepth]
        return cmd


class HM(VTM):
    """HM: HEVC reference software"""

    binext = "bin"
    config = ""

    @property
    def name(self):
        return "HM"

    @property
    def description(self):
        return f"HM reference software, version {self.get_version(self.encoder_path)}"

    def name_config(self):
        return f"{self.name}-v{self.get_version(self.encoder_path)}-{self.config}"

    def get_encoder_path(self, build_dir):
        system = platform.system()
        try:
            elfnames = {"Darwin": "TAppEncoder", "Linux": "TAppEncoderStatic"}
            return Path(build_dir) / elfnames[system]
        except KeyError as err:
            raise RuntimeError(f'Unsupported platform "{system}"') from err

    def get_decoder_path(self, build_dir):
        system = platform.system()
        try:
            elfnames = {"Darwin": "TAppDecoder", "Linux": "TAppDecoderStatic"}
            return Path(build_dir) / elfnames[system]
        except KeyError as err:
            raise RuntimeError(f'Unsupported platform "{system}"') from err

    def set_args(self, args):
        args = super().set_args(args)
        self.encoder_path = self.get_encoder_path(args.build_dir)
        self.decoder_path = self.get_decoder_path(args.build_dir)
        self.config_path = args.config
        self.config = Path(self.config_path).stem.split("_")[1]
        self.version = self.get_version(self.encoder_path)
        self.rgb = args.rgb
        return args

    def get_encode_cmd(self, filepath: Path, qp, binpath) -> List[Any]:
        info = get_raw_video_file_info(filepath.stem)
        num_frames = len(RawVideoSequence.from_file(str(filepath)))
        cmd = [
            self.encoder_path,
            "-i",
            filepath,
            "-c",
            self.config_path,
            "-q",
            qp,
            "-o",
            "/dev/null",
            "-b",
            binpath,
            "-wdt",
            info["width"],
            "-hgt",
            info["height"],
            "-fr",
            info["framerate"],
            "-f",
            num_frames,
            f'--InputBitDepth={info["bitdepth"]}',
            f'--OutputBitDepth={info["bitdepth"]}',
            # "--ConformanceWindowMode=1",
        ]

        if self.rgb:
            cmd += [
                "--InputColourSpaceConvert=RGBtoGBR",
                "--SNRInternalColourSpace=1",
                "--OutputInternalColourSpace=0",
            ]
        return cmd

    def get_decode_cmd(
        self, binpath: Path, decpath: Path, input_filepath: Path
    ) -> List[Any]:
        output_bitdepth = get_raw_video_file_info(input_filepath.stem)["bitdepth"]
        cmd = [self.decoder_path, "-b", binpath, "-o", decpath, "-d", output_bitdepth]
        return cmd



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Code adapted from https://github.com/yunhe20/D-PCC

import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange, repeat

try:
    from pointops.functions import pointops
except ImportError:
    pass  # NOTE: Optional dependency.

from .utils import index_points


class DownsampleLayer(nn.Module):
    """Downsampling layer used in [He2022pcc]_.

    Downsamples positions into a smaller number of centroids.
    Each centroid is grouped with nearby points,
    and the local point density is estimated for that group.
    Then, the positions, features, and density for the group
    are embedded into a single aggregate vector from which the
    group of points may later be reconstructed.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, downsample_rate, dim, hidden_dim, k, ngroups):
        super().__init__()
        self.k = k
        self.downsample_rate = downsample_rate
        self.pre_conv = nn.Conv1d(dim, dim, 1)
        self.embed_features = PointTransformerLayer(dim, dim, hidden_dim, ngroups)
        self.embed_positions = PositionEmbeddingLayer(hidden_dim, dim, ngroups)
        self.embed_densities = DensityEmbeddingLayer(hidden_dim, dim, ngroups)
        self.post_conv = nn.Conv1d(dim * 3, dim, 1)

    def get_density(self, downsampled_xyzs, input_xyzs):
        # downsampled_xyzs: (b, 3, m)
        # input_xyzs: (b, 3, n)
        # nn_idx: (b, n, 1)
        # downsample_num: (b, m)
        # knn_idx: (b, m, k)
        # mask: (b, m, k)
        # distance: (b, m)
        # mean_distance: (b, m)
        _, _, n = input_xyzs.shape
        distance, mask, knn_idx, _ = nearby_distance_sum(
            downsampled_xyzs, input_xyzs, min(self.k, n)
        )
        downsample_num = mask.sum(dim=-1).float()
        mean_distance = distance / downsample_num
        return downsample_num, mean_distance, mask, knn_idx

    def forward(self, xyzs, feats):
        # xyzs: (b, 3, n)
        # features: (b, cin, n)
        # sample_idx: (b, m)
        # sampled_xyzs: (b, 3, m)
        # sampled_feats: (b, c, m)

        # Downsample positions into a smaller number of centroids.
        sampled_xyzs, sample_idx = self.downsample_positions(xyzs, feats)

        # For each centroid, form a group with nearby points.
        # Also, estimate local point density ("mean distance") for each centroid.
        downsample_num, mean_distance, mask, knn_idx = self.get_density(
            sampled_xyzs, xyzs
        )

        # Embed features, positions, and density for each downsampled
        # point group into a single aggregate vector for that group.
        sampled_feats = self.downsample_features(
            sampled_xyzs, xyzs, feats, downsample_num, sample_idx, knn_idx, mask
        )

        return sampled_xyzs, sampled_feats, downsample_num, mean_distance

    def downsample_positions(self, xyzs, sample_num):
        _, _, n = xyzs.shape
        sample_num = round(n * self.downsample_rate)
        xyzs_tr = xyzs.permute(0, 2, 1).contiguous()
        sample_idx = pointops.furthestsampling(xyzs_tr, sample_num).long()
        sampled_xyzs = index_points(xyzs, sample_idx)
        return sampled_xyzs, sample_idx

    def downsample_features(
        self, sampled_xyzs, xyzs, feats, downsample_num, sample_idx, knn_idx, mask
    ):
        # sampled_xyzs: (b, 3, m)
        # sampled_feats: (b, c, m)

        identity = index_points(feats, sample_idx)

        feats = self.pre_conv(feats)
        sampled_feats = index_points(feats, sample_idx)
        embeddings = [
            self.embed_features(
                sampled_xyzs, xyzs, sampled_feats, feats, feats, knn_idx, mask
            ),
            self.embed_positions(sampled_xyzs, xyzs, knn_idx, mask),
            self.embed_densities(downsample_num.unsqueeze(1)),
        ]
        agg_embedding = self.post_conv(torch.cat(embeddings, dim=1))

        sampled_feats_new = agg_embedding + identity
        return sampled_feats_new


class PointTransformerLayer(nn.Module):
    """Point Transformer layer introduced by [Zhao2021]_.

    References:

        .. [Zhao2021] `"Point Transformer"
            <https://arxiv.org/abs/2012.09164>`_, by Hengshuang Zhao,
            Li Jiang, Jiaya Jia, Philip Torr, and Vladlen Koltun,
            CVPR 2021.
    """

    def __init__(self, in_fdim, out_fdim, hidden_dim, ngroups):
        super().__init__()

        self.w_qs = nn.Conv1d(in_fdim, hidden_dim, 1)
        self.w_ks = nn.Conv1d(in_fdim, hidden_dim, 1)
        self.w_vs = nn.Conv1d(in_fdim, hidden_dim, 1)

        self.conv_delta = nn.Sequential(
            nn.Conv2d(3, hidden_dim, 1),
            nn.GroupNorm(ngroups, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Conv2d(hidden_dim, hidden_dim, 1),
        )

        self.conv_gamma = nn.Sequential(
            nn.Conv2d(hidden_dim, hidden_dim, 1),
            nn.GroupNorm(ngroups, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Conv2d(hidden_dim, hidden_dim, 1),
        )

        self.post_conv = nn.Conv1d(hidden_dim, out_fdim, 1)

    def forward(self, q_xyzs, k_xyzs, q_feats, k_feats, v_feats, knn_idx, mask):
        # q: (b, c, m)
        # k: (b, c, n)
        # knn_idx: (b, m, k)
        # mask: (b, m, k)
        # knn_xyzs: (b, 3, m, k)
        # query: (b, c, m)
        # key: (b, c, m, k)
        # pos_enc: (b, c, m, k)
        # attn: (b, c, m, k)

        knn_xyzs = index_points(k_xyzs, knn_idx)

        # NOTE: it's q_feats, not v_feats!
        identity = q_feats

        query = self.w_qs(q_feats)
        key = index_points(self.w_ks(k_feats), knn_idx)
        value = index_points(self.w_vs(v_feats), knn_idx)

        pos_enc = self.conv_delta(q_xyzs.unsqueeze(-1) - knn_xyzs)

        attn = self.conv_gamma(query.unsqueeze(-1) - key + pos_enc)
        attn = attn / math.sqrt(key.shape[1])
        mask_value = -(torch.finfo(attn.dtype).max)
        attn.masked_fill_(~mask[:, None], mask_value)
        attn = F.softmax(attn, dim=-1)

        result = torch.einsum("bcmk, bcmk -> bcm", attn, value + pos_enc)
        result = self.post_conv(result) + identity

        return result


class PositionEmbeddingLayer(nn.Module):
    """Position embedding for downsampling, as introduced in [He2022pcc]_.

    For each group of feature vectors (f₁, ..., fₖ) with centroid fₒ,
    represents the offsets (f₁ - fₒ, ..., fₖ - fₒ) as
    magnitude-direction vectors, then applies an MLP to each vector,
    then takes a softmax self-attention over the resulting vectors,
    and finally reduces the vectors via a sum,
    resulting in a single embedded vector for the group.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, hidden_dim, dim, ngroups):
        super().__init__()

        self.embed_positions = nn.Sequential(
            nn.Conv2d(4, hidden_dim, 1),
            nn.GroupNorm(ngroups, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Conv2d(hidden_dim, dim, 1),
        )

        self.attention = nn.Sequential(
            nn.Conv2d(dim, hidden_dim, 1),
            nn.GroupNorm(ngroups, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Conv2d(hidden_dim, dim, 1),
        )

    def forward(self, q_xyzs, k_xyzs, knn_idx, mask):
        # q_xyzs: (b, 3, m)
        # k_xyzs: (b, 3, n)
        # knn_idx: (b, m, k)
        # mask: (b, m, k)
        # knn_xyzs: (b, 3, m, k)
        # repeated_xyzs: (b, 3, m, k)
        # direction: (b, 3, m, k)
        # distance: (b, 1, m, k)
        # local_pattern: (b, 4, m, k)
        # position_embedding_expanded: (b, c, m, k)
        # attn: (b, c, m, k)
        # position_embedding: (b, c, m)

        # "query" (q_xyzs) points are the centroids of each point group.
        # "key" (k_xyzs) points are points in the neighborhood of each centroid.

        _, _, k = knn_idx.shape
        knn_xyzs = index_points(k_xyzs, knn_idx)
        repeated_xyzs = q_xyzs[..., None].repeat(1, 1, 1, k)

        # Represent points within a point group as (direction, distance)
        # of offsets from the group centroid.
        offset_xyzs = knn_xyzs - repeated_xyzs
        direction = F.normalize(offset_xyzs, p=2, dim=1)
        distance = torch.linalg.norm(offset_xyzs, dim=1, keepdim=True)
        local_pattern = torch.cat((direction, distance), dim=1)

        # Apply a pointwise MLP to each point.
        position_embedding_expanded = self.embed_positions(local_pattern)

        # Compute self-attention, ignoring points that are not in the
        # neighborhood of the centroid.
        attn = self.attention(position_embedding_expanded)
        mask_value = -(torch.finfo(attn.dtype).max)
        attn.masked_fill_(~mask[:, None], mask_value)
        attn = F.softmax(attn, dim=-1)
        position_embedding = (position_embedding_expanded * attn).sum(dim=-1)

        return position_embedding


class DensityEmbeddingLayer(nn.Module):
    """Density embedding for downsampling, as introduced in [He2022pcc]_.

    Applies an embedding ℝ → ℝᶜ to the local point density (scalar).
    The local point density is measured using the mean distance of the
    points within the neighborhood of a "downsampled" centroid.
    This information is useful when upsampling from the single centroid.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, hidden_dim, dim, ngroups):
        super().__init__()
        self.embed_densities = nn.Sequential(
            nn.Conv1d(1, hidden_dim, 1),
            nn.GroupNorm(ngroups, hidden_dim),
            nn.ReLU(inplace=True),
            nn.Conv1d(hidden_dim, dim, 1),
        )

    def forward(self, downsample_num):
        # downsample_num: (b, 1, n)
        # density_embedding: (b, c, n)
        density_embedding = self.embed_densities(downsample_num)
        return density_embedding


class UpsampleLayer(nn.Module):
    """Upsampling layer used in [He2022pcc]_.

    Upsamples many candidate points from a smaller number of centroids.
    (Not all candidate upsampled points will be kept; some will be
    thrown away to match the predicted local point density.)

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, dim, hidden_dim, k, sub_point_conv_mode, upsample_rate):
        super().__init__()
        self.xyzs_upsample_nn = XyzsUpsampleLayer(
            dim, hidden_dim, k, sub_point_conv_mode, upsample_rate
        )
        self.feats_upsample_nn = FeatsUpsampleLayer(
            dim, hidden_dim, k, sub_point_conv_mode, upsample_rate
        )

    def forward(self, xyzs, feats):
        upsampled_xyzs = self.xyzs_upsample_nn(xyzs, feats)
        upsampled_feats = self.feats_upsample_nn(feats)
        return upsampled_xyzs, upsampled_feats


class UpsampleNumLayer(nn.Module):
    """Predicts local point density while upsampling, as used in [He2022pcc]_.

    Extracts the number of candidate points to keep after upsampling
    from a given "centroid" feature vector.
    (Some candidate upsampled points will be thrown away to match the
    predicted local point density.)

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, dim, hidden_dim, upsample_rate):
        super().__init__()
        self.upsample_rate = upsample_rate
        self.upsample_num_nn = nn.Sequential(
            nn.Conv1d(dim, hidden_dim, 1),
            nn.ReLU(),
            nn.Conv1d(hidden_dim, 1, 1),
            nn.Sigmoid(),
        )

    def forward(self, feats):
        # upsample_num: (b, n)
        upsample_frac = self.upsample_num_nn(feats).squeeze(1)
        upsample_num = upsample_frac * (self.upsample_rate - 1) + 1
        return upsample_num


class RefineLayer(nn.Module):
    """Refines upsampled points, as used in [He2022pcc]_.

    After the centroids are upsampled, there may be overlapping
    point groups between nearby centroids, and other artifacts.
    Refinement should help correct various such artifacts.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, dim, hidden_dim, k, sub_point_conv_mode, decompress_normal):
        super().__init__()

        self.xyzs_refine_nn = XyzsUpsampleLayer(
            dim,
            hidden_dim,
            k,
            sub_point_conv_mode,
            upsample_rate=1,
        )

        self.feats_refine_nn = FeatsUpsampleLayer(
            dim,
            hidden_dim,
            k,
            sub_point_conv_mode,
            upsample_rate=1,
            decompress_normal=decompress_normal,
        )

    def forward(self, xyzs, feats):
        # refined_xyzs: (b, 3, n, 1)
        # refined_xyzs: (b, 3, n)  [after rearrange]
        # refined_feats: (b, c, n, 1)
        # refined_feats: (b, c, n)  [after rearrange]

        refined_xyzs = self.xyzs_refine_nn(xyzs, feats)
        refined_xyzs = rearrange(refined_xyzs, "b c n u -> b c (n u)")

        refined_feats = self.feats_refine_nn(feats)
        refined_feats = rearrange(refined_feats, "b c n u -> b c (n u)")

        return refined_xyzs, refined_feats


class XyzsUpsampleLayer(nn.Module):
    """Position upsampling layer used in [He2022pcc]_.

    Upsamples many positions from each "centroid" feature vector.
    Each feature vector is upsampled into various offsets represented as
    magnitude-direction vectors, where each direction is determined by a
    weighted sum of various fixed hypothesized directions.
    From this, the candidate upsampled positions are simply the
    the offset vectors plus their original centroid position.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, dim, hidden_dim, k, sub_point_conv_mode, upsample_rate):
        super().__init__()

        self.upsample_rate = upsample_rate

        # The hypothesis is a basis of 43 candidate direction vectors.
        hypothesis, _ = icosahedron2sphere(1)
        hypothesis = np.append(np.zeros((1, 3)), hypothesis, axis=0)
        self.hypothesis = torch.from_numpy(hypothesis).float().cuda()

        self.weight_nn = SubPointConv(
            hidden_dim, k, sub_point_conv_mode, dim, 43 * upsample_rate, upsample_rate
        )

        self.scale_nn = SubPointConv(
            hidden_dim, k, sub_point_conv_mode, dim, 1 * upsample_rate, upsample_rate
        )

    def forward(self, xyzs, feats):
        # xyzs: (b, 3, n)
        # feats: (b, c, n)
        # weights: (b, 43, n, u)
        # weights: (b, 43, 1, n, u)  [after unsqueeze]
        # hypothesis: (b, 43, 3, n, u)
        # directions: (b, 3, n, u)
        # scales: (b, 1, n, u)
        # deltas: (b, 3, n, u)
        # repeated_xyzs: (b, 3, n, u)

        batch_size = xyzs.shape[0]
        points_num = xyzs.shape[2]

        weights = self.weight_nn(feats)
        weights = weights.unsqueeze(2)
        weights = F.softmax(weights, dim=1)

        hypothesis = repeat(
            self.hypothesis,
            "h c -> b h c n u",
            b=batch_size,
            n=points_num,
            u=self.upsample_rate,
        )
        weighted_hypothesis = weights * hypothesis
        directions = torch.sum(weighted_hypothesis, dim=1)
        directions = F.normalize(directions, p=2, dim=1)

        scales = self.scale_nn(feats)

        deltas = directions * scales

        repeated_xyzs = repeat(xyzs, "b c n -> b c n u", u=self.upsample_rate)
        upsampled_xyzs = repeated_xyzs + deltas

        return upsampled_xyzs


class FeatsUpsampleLayer(nn.Module):
    """Feature upsampling layer used in [He2022pcc]_.

    Upsamples many features from each "centroid" feature vector.
    The feature vector associated with each centroid is upsampled
    into various candidate upsampled feature vectors.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(
        self,
        dim,
        hidden_dim,
        k,
        sub_point_conv_mode,
        upsample_rate,
        decompress_normal=False,
    ):
        super().__init__()

        self.upsample_rate = upsample_rate
        self.decompress_normal = decompress_normal
        out_fdim = (3 if decompress_normal else dim) * upsample_rate

        self.feats_nn = SubPointConv(
            hidden_dim, k, sub_point_conv_mode, dim, out_fdim, upsample_rate
        )

    def forward(self, feats):
        # upsampled_feats: (b, c, n, u)
        upsampled_feats = self.feats_nn(feats)
        if not self.decompress_normal:
            repeated_feats = repeat(feats, "b c n -> b c n u", u=self.upsample_rate)
            upsampled_feats = upsampled_feats + repeated_feats
        return upsampled_feats


class SubPointConv(nn.Module):
    """Sub-point convolution for upsampling, as introduced in [He2022pcc]_.

    Each feature vector (representing a "centroid" point) is sliced
    into g feature vectors, where each feature vector represents a
    point that has been upsampled from the original centroid point.
    Then, an MLP is applied to each slice individually.

    References:

        .. [He2022pcc] `"Density-preserving Deep Point Cloud Compression"
            <https://arxiv.org/abs/2204.12684>`_, by Yun He, Xinlin Ren,
            Danhang Tang, Yinda Zhang, Xiangyang Xue, and Yanwei Fu,
            CVPR 2022.
    """

    def __init__(self, hidden_dim, k, mode, in_fdim, out_fdim, group_num):
        super().__init__()

        self.mode = mode
        self.group_num = group_num
        group_in_fdim = in_fdim // group_num
        group_out_fdim = out_fdim // group_num

        if self.mode == "mlp":
            self.mlp = nn.Sequential(
                nn.Conv2d(group_in_fdim, hidden_dim, 1),
                nn.ReLU(inplace=True),
                nn.Conv2d(hidden_dim, group_out_fdim, 1),
            )
        elif self.mode == "edge_conv":
            self.edge_conv = EdgeConv(in_fdim, out_fdim, hidden_dim, k)
        else:
            raise ValueError(f"Unknown mode: {self.mode}")

    def forward(self, feats):
        # feats: (b, cin * g, n)
        # expanded_feats: (b, cout, n, g)

        if self.mode == "mlp":
            feats = rearrange(
                feats, "b (c g) n -> b c n g", g=self.group_num
            ).contiguous()
            expanded_feats = self.mlp(feats)
        elif self.mode == "edge_conv":
            expanded_feats = self.edge_conv(feats)
            expanded_feats = rearrange(
                expanded_feats, "b (c g) n -> b c n g", g=self.group_num
            ).contiguous()
        else:
            raise ValueError(f"Unknown mode: {self.mode}")

        return expanded_feats


class EdgeConv(nn.Module):
    """EdgeConv introduced by [Wang2019dgcnn]_.

    First, groups similar feature vectors together via k-nearest neighbors
    using the following distance metric between feature vectors fᵢ and fⱼ:
    distance[i, j] = 2fᵢᵀfⱼ - ||fᵢ||² - ||fⱼ||².

    Then, for each group of feature vectors (f₁, ..., fₖ) with centroid fₒ,
    the residual feature vectors are each concatenated with the centroid,
    then an MLP is applied to each resulting vector individually,
    i.e., (MLP(f₁ - fₒ, fₒ), ..., MLP(fₖ - fₒ, fₒ)),
    and finally the elementwise max is taken across the resulting vectors,
    resulting in a single vector fₘₐₓ for the group.

    Original code located at [DGCNN]_ under MIT License.

    References:

        .. [Wang2019dgcnn] `"Dynamic Graph CNN for Learning on Point Clouds"
            <https://arxiv.org/abs/1801.07829>`_, by Yue Wang, Yongbin
            Sun, Ziwei Liu, Sanjay E. Sarma, Michael M. Bronstein,
            Justin M. Solomon, ACM Transactions on Graphics 2019.

        .. [DGCNN] `DGCNN
            <https://github.com/WangYueFt/dgcnn/blob/master/pytorch/model.py>`_
    """

    def __init__(self, in_fdim, out_fdim, hidden_dim, k):
        super().__init__()
        self.k = k
        self.conv = nn.Sequential(
            nn.Conv2d(2 * in_fdim, hidden_dim, 1),
            nn.ReLU(),
            nn.Conv2d(hidden_dim, out_fdim, 1),
        )

    # WARN: This requires at least O(n^2) memory.
    def knn(self, feats, k):
        # feats: (b, c, n)
        # sq_norm: (b, 1, n)
        # pairwise_dot: (b, n, n)
        # pairwise_distance: (b, n, n)
        # knn_idx: (b, n, k)

        sq_norm = (feats**2).sum(dim=1, keepdim=True)

        # Pairwise dot product fᵢᵀfⱼ between feature vectors fᵢ and fⱼ.
        pairwise_dot = torch.matmul(feats.transpose(2, 1), feats)

        # pairwise_distance[i, j] = 2fᵢᵀfⱼ - ||fᵢ||² - ||fⱼ||²
        pairwise_distance = 2 * pairwise_dot - sq_norm - sq_norm.transpose(2, 1)

        _, knn_idx = pairwise_distance.topk(k=k, dim=-1)
        return knn_idx

    def get_graph_features(self, feats, k):
        # knn_feats: (b, c, n, k)
        # graph_feats: (b, 2c, n, k)
        dim = feats.shape[1]
        if dim == 3:
            feats_tr = feats.permute(0, 2, 1).contiguous()
            knn_idx = pointops.knnquery_heap(k, feats_tr, feats_tr).long()
        else:
            knn_idx = self.knn(feats, k)
        torch.cuda.empty_cache()
        knn_feats = index_points(feats, knn_idx)
        repeated_feats = repeat(feats, "b c n -> b c n k", k=k)
        graph_feats = torch.cat((knn_feats - repeated_feats, repeated_feats), dim=1)
        return graph_feats

    def forward(self, feats):
        # feats: (b, c, n)
        # graph_feats: (b, 2c, n, k)
        # expanded_feats: (b, cout*g, n, k)
        # feats_new: (b, cout*g, n)
        _, _, n = feats.shape
        graph_feats = self.get_graph_features(feats, k=min(self.k, n))
        expanded_feats = self.conv(graph_feats)
        feats_new, _ = expanded_feats.max(dim=-1)
        return feats_new


def icosahedron2sphere(level):
    """Samples uniformly on a sphere using a icosahedron.

    Code adapted from [IcoSphere_MATLAB]_ and [IcoSphere_Python]_,
    from paper [Xiao2009]_.

    References:

        .. [Xiao2009] `"Image-based street-side city modeling"
            <https://dl.acm.org/doi/10.1145/1618452.1618460>`_,
            by Jianxiong Xiao, Tian Fang, Peng Zhao, Maxime Lhuillier,
            and Long Quan, ACM Transactions on Graphics, 2009.

        .. [IcoSphere_MATLAB] https://github.com/jianxiongxiao/ProfXkit/blob/master/icosahedron2sphere/icosahedron2sphere.m

        .. [IcoSphere_Python] https://github.com/23michael45/PanoContextTensorflow/blob/master/PanoContextTensorflow/icosahedron2sphere.py
    """
    a = 2 / (1 + np.sqrt(5))

    # fmt: off
    M = np.array([
         0,  a, -1,  a,  1,  0, -a,  1,  0,  # noqa: E241, E126
         0,  a,  1, -a,  1,  0,  a,  1,  0,  # noqa: E241
         0,  a,  1,  0, -a,  1, -1,  0,  a,  # noqa: E241
         0,  a,  1,  1,  0,  a,  0, -a,  1,  # noqa: E241
         0,  a, -1,  0, -a, -1,  1,  0, -a,  # noqa: E241
         0,  a, -1, -1,  0, -a,  0, -a, -1,  # noqa: E241
         0, -a,  1,  a, -1,  0, -a, -1,  0,  # noqa: E241
         0, -a, -1, -a, -1,  0,  a, -1,  0,  # noqa: E241
        -a,  1,  0, -1,  0,  a, -1,  0, -a,  # noqa: E241, E131
        -a, -1,  0, -1,  0, -a, -1,  0,  a,  # noqa: E241
         a,  1,  0,  1,  0, -a,  1,  0,  a,  # noqa: E241
         a, -1,  0,  1,  0,  a,  1,  0, -a,  # noqa: E241
         0,  a,  1, -1,  0,  a, -a,  1,  0,  # noqa: E241
         0,  a,  1,  a,  1,  0,  1,  0,  a,  # noqa: E241
         0,  a, -1, -a,  1,  0, -1,  0, -a,  # noqa: E241
         0,  a, -1,  1,  0, -a,  a,  1,  0,  # noqa: E241
         0, -a, -1, -1,  0, -a, -a, -1,  0,  # noqa: E241
         0, -a, -1,  a, -1,  0,  1,  0, -a,  # noqa: E241
         0, -a,  1, -a, -1,  0, -1,  0,  a,  # noqa: E241
         0, -a,  1,  1,  0,  a,  a, -1,  0,  # noqa: E241
    ])
    # fmt: on

    coor = M.reshape(60, 3)
    coor, idx = np.unique(coor, return_inverse=True, axis=0)
    tri = idx.reshape(20, 3)

    # extrude
    coor_norm = np.linalg.norm(coor, axis=1, keepdims=True)
    coor = list(coor / np.tile(coor_norm, (1, 3)))

    for _ in range(level):
        tris = []

        for t in range(len(tri)):
            n = len(coor)
            coor.extend(
                [
                    (coor[tri[t, 0]] + coor[tri[t, 1]]) / 2,
                    (coor[tri[t, 1]] + coor[tri[t, 2]]) / 2,
                    (coor[tri[t, 2]] + coor[tri[t, 0]]) / 2,
                ]
            )
            tris.extend(
                [
                    [n, tri[t, 0], n + 2],
                    [n, tri[t, 1], n + 1],
                    [n + 1, tri[t, 2], n + 2],
                    [n, n + 1, n + 2],
                ]
            )

        tri = np.array(tris)

        # uniquefy
        coor, idx = np.unique(coor, return_inverse=True, axis=0)
        tri = idx[tri]

        # extrude
        coor_norm = np.linalg.norm(coor, axis=1, keepdims=True)
        coor = list(coor / np.tile(coor_norm, (1, 3)))

    return np.array(coor), np.array(tri)


def nearby_distance_sum(a_xyzs, b_xyzs, k):
    """Computes sum of nearby distances to B for each point in A.

    Partitions a point set B into non-intersecting sets
    C(a_1), ..., C(a_m) where each C(a_i) contains points that are
    nearest to a_i ∈ A.
    For each a_i ∈ A, computes the total distance from a_i to C(a_i).
    (Note that C(a_1), ..., C(a_m) may not cover all of B.)

    In more precise terms:
    For each a ∈ A, let C(a) ⊆ B denote its "collapsed point set" s.t.
    (i)   b ∈ C(a)  ⇒  min_{a' ∈ A} ||a' - b|| = ||a - b||,
    (ii)  ⋃ _{a ∈ A} C(a) ⊆ B,
    (iii) ⋂ _{a ∈ A} C(a) = ∅, and
    (iv)  |C(a)| ≤ k.
    For each a ∈ A, we then compute d(a) = ∑_{b ∈ C(a)} ||a - b||.

    Args:
        a_xyzs: (b, 3, m) Input point set A.
        b_xyzs: (b, 3, n) Input point set B.
        k: Maximum number of points in each collapsed point set C(a_i).

    Returns:
        distance: (b, m) Sum of distances from each point in A to its
            collapsed point set.
        mask: (b, m, k) Mask indicating which points in the ``knn_idx``
            belong to the collapsed point set of each point in A.
        knn_idx: (b, m, k) Indices of the points in B that are nearest
            to each point in A.
        nn_idx: (b, n, 1) Indices of the point in A that is nearest
            to each point in B.
    """
    # expect_idx: (b, m, k)
    # actual_idx: (b, m, k)
    # knn_xyzs: (b, 3, m, k)
    # knn_distances: (b, m, k)

    device = a_xyzs.device
    _, _, m = a_xyzs.shape
    a_xyzs_tr = a_xyzs.permute(0, 2, 1).contiguous()
    b_xyzs_tr = b_xyzs.permute(0, 2, 1).contiguous()

    # Determine which point in A each point in B is closest to.
    nn_idx = pointops.knnquery_heap(1, a_xyzs_tr, b_xyzs_tr)
    nn_idx_tr = nn_idx.permute(0, 2, 1).contiguous()

    # Determine k nearest neighbors in B for each point in A.
    knn_idx = pointops.knnquery_heap(k, b_xyzs_tr, a_xyzs_tr).long()
    torch.cuda.empty_cache()

    # Mask points that do not belong to the collapsed points set C(a).
    expect_idx = torch.arange(m, device=device)[None, :, None]
    actual_idx = index_points(nn_idx_tr, knn_idx).squeeze(1)
    mask = expect_idx == actual_idx

    # Compute the distance from each A point to its k nearest neighbors in B.
    knn_xyzs = index_points(b_xyzs, knn_idx)
    knn_distances = torch.linalg.norm(knn_xyzs - a_xyzs[..., None], dim=1)

    # Zero away the distances for points that are not in C(a).
    # knn_distances.masked_fill_(~mask, 0)
    knn_distances = knn_distances * mask.float()

    # Compute masked distances.
    # Notably, distance.sum(-1) is upper bounded by the sum of the
    # distances from each point in B to its nearest point in A.
    distance = knn_distances.sum(dim=-1)

    return distance, mask, knn_idx, nn_idx



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Based on https://github.com/facebookresearch/ParlAI/tree/c06c40603f45918f58cb09122fa8c74dd4047057/docs/source

import importlib
import io

from pathlib import Path

import compressai.utils


def get_utils():
    rootdir = Path(compressai.utils.__file__).parent
    for d in sorted(rootdir.iterdir()):
        if d.is_dir() and (d / "__main__.py").is_file():
            yield d


def main():
    fout = open("cli_usage.inc", "w")

    for p in get_utils():
        try:
            m = importlib.import_module(f"compressai.utils.{p.name}.__main__")
        except ImportError:
            continue

        if not hasattr(m, "setup_args"):
            continue

        fout.write(p.name)
        fout.write("\n")
        fout.write("-" * len(p.name))
        fout.write("\n")

        doc = m.__doc__
        if doc:
            fout.write(doc)
            fout.write("\n")

        fout.write(".. code-block:: text\n\n")
        capture = io.StringIO()
        parser = m.setup_args()
        if isinstance(parser, tuple):
            parser = parser[0]
        parser.prog = f"python -m compressai.utils.{p.name}"
        parser.print_help(capture)

        for line in capture.getvalue().split("\n"):
            fout.write(f"\t{line}\n")

        fout.write("\n\n")

    fout.close()


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

from PIL import Image
from torch.utils.data import Dataset

from compressai.registry import register_dataset


@register_dataset("ImageFolder")
class ImageFolder(Dataset):
    """Load an image folder database. Training and testing image samples
    are respectively stored in separate directories:

    .. code-block::

        - rootdir/
            - train/
                - img000.png
                - img001.png
            - test/
                - img000.png
                - img001.png

    Args:
        root (string): root directory of the dataset
        transform (callable, optional): a function or transform that takes in a
            PIL image and returns a transformed version
        split (string): split mode ('train' or 'val')
    """

    def __init__(self, root, transform=None, split="train"):
        splitdir = Path(root) / split

        if not splitdir.is_dir():
            raise RuntimeError(f'Missing directory "{splitdir}"')

        self.samples = sorted(f for f in splitdir.iterdir() if f.is_file())

        self.transform = transform

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            img: `PIL.Image.Image` or transformed `PIL.Image.Image`.
        """
        img = Image.open(self.samples[index]).convert("RGB")
        if self.transform:
            return self.transform(img)
        return img

    def __len__(self):
        return len(self.samples)



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Dict, List, Optional, Tuple, Union

import torch.nn as nn

from torch import Tensor

from compressai.entropy_models import GaussianConditional
from compressai.ops import quantize_ste
from compressai.registry import register_module

from .base import LatentCodec

__all__ = [
    "GaussianConditionalLatentCodec",
]


@register_module("GaussianConditionalLatentCodec")
class GaussianConditionalLatentCodec(LatentCodec):
    """Gaussian conditional for compressing latent ``y`` using ``ctx_params``.

    Probability model for Gaussian of ``(scales, means)``.

    Gaussian conditonal entropy model introduced in
    `"Variational Image Compression with a Scale Hyperprior"
    <https://arxiv.org/abs/1802.01436>`_,
    by J. Balle, D. Minnen, S. Singh, S.J. Hwang, and N. Johnston,
    International Conference on Learning Representations (ICLR), 2018.

    .. note:: Unlike the original paper, which models only the scale
       (i.e. "width") of the Gaussian, this implementation models both
       the scale and the mean (i.e. "center") of the Gaussian.

    .. code-block:: none

                          ctx_params
                              │
                              ▼
                              │
                           ┌──┴──┐
                           │  EP │
                           └──┬──┘
                              │
               ┌───┐  y_hat   ▼
        y ──►──┤ Q ├────►────····──►── y_hat
               └───┘          GC

    """

    gaussian_conditional: GaussianConditional
    entropy_parameters: nn.Module

    def __init__(
        self,
        scale_table: Optional[Union[List, Tuple]] = None,
        gaussian_conditional: Optional[GaussianConditional] = None,
        entropy_parameters: Optional[nn.Module] = None,
        quantizer: str = "noise",
        chunks: Tuple[str, ...] = ("scales", "means"),
        **kwargs,
    ):
        super().__init__()
        self.quantizer = quantizer
        self.gaussian_conditional = gaussian_conditional or GaussianConditional(
            scale_table, **kwargs
        )
        self.entropy_parameters = entropy_parameters or nn.Identity()
        self.chunks = tuple(chunks)

    def forward(self, y: Tensor, ctx_params: Tensor) -> Dict[str, Any]:
        gaussian_params = self.entropy_parameters(ctx_params)
        scales_hat, means_hat = self._chunk(gaussian_params)
        y_hat, y_likelihoods = self.gaussian_conditional(y, scales_hat, means=means_hat)
        if self.quantizer == "ste":
            y_hat = quantize_ste(y - means_hat) + means_hat
        return {"likelihoods": {"y": y_likelihoods}, "y_hat": y_hat}

    def compress(self, y: Tensor, ctx_params: Tensor) -> Dict[str, Any]:
        gaussian_params = self.entropy_parameters(ctx_params)
        scales_hat, means_hat = self._chunk(gaussian_params)
        indexes = self.gaussian_conditional.build_indexes(scales_hat)
        y_strings = self.gaussian_conditional.compress(y, indexes, means_hat)
        y_hat = self.gaussian_conditional.decompress(
            y_strings, indexes, means=means_hat
        )
        return {"strings": [y_strings], "shape": y.shape[2:4], "y_hat": y_hat}

    def decompress(
        self,
        strings: List[List[bytes]],
        shape: Tuple[int, int],
        ctx_params: Tensor,
        **kwargs,
    ) -> Dict[str, Any]:
        (y_strings,) = strings
        gaussian_params = self.entropy_parameters(ctx_params)
        scales_hat, means_hat = self._chunk(gaussian_params)
        indexes = self.gaussian_conditional.build_indexes(scales_hat)
        y_hat = self.gaussian_conditional.decompress(
            y_strings, indexes, means=means_hat
        )
        assert y_hat.shape[2:4] == shape
        return {"y_hat": y_hat}

    def _chunk(self, params: Tensor) -> Tuple[Tensor, Tensor]:
        scales, means = None, None
        if self.chunks == ("scales",):
            scales = params
        if self.chunks == ("means",):
            means = params
        if self.chunks == ("scales", "means"):
            scales, means = params.chunk(2, 1)
        if self.chunks == ("means", "scales"):
            means, scales = params.chunk(2, 1)
        return scales, means



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from itertools import accumulate
from typing import Any, Dict, List, Mapping, Tuple

import torch
import torch.nn as nn

from torch import Tensor

from compressai.registry import register_module

from .base import LatentCodec

__all__ = [
    "ChannelGroupsLatentCodec",
]


@register_module("ChannelGroupsLatentCodec")
class ChannelGroupsLatentCodec(LatentCodec):
    """Reconstructs groups of channels using previously decoded groups.

    Context model from [Minnen2020] and [He2022].
    Also known as a "channel-conditional" (CC) entropy model.

    See :py:class:`~compressai.models.sensetime.Elic2022Official`
    for example usage.

    [Minnen2020]: `"Channel-wise Autoregressive Entropy Models for
    Learned Image Compression" <https://arxiv.org/abs/2007.08739>`_, by
    David Minnen, and Saurabh Singh, ICIP 2020.

    [He2022]: `"ELIC: Efficient Learned Image Compression with
    Unevenly Grouped Space-Channel Contextual Adaptive Coding"
    <https://arxiv.org/abs/2203.10886>`_, by Dailan He, Ziming Yang,
    Weikun Peng, Rui Ma, Hongwei Qin, and Yan Wang, CVPR 2022.
    """

    latent_codec: Mapping[str, LatentCodec]

    channel_context: Mapping[str, nn.Module]

    def __init__(
        self,
        latent_codec: Mapping[str, LatentCodec],
        channel_context: Mapping[str, nn.Module],
        *,
        groups: List[int],
        **kwargs,
    ):
        super().__init__()
        self._kwargs = kwargs
        self.groups = list(groups)
        self.groups_acc = list(accumulate(self.groups, initial=0))
        self.channel_context = nn.ModuleDict(channel_context)
        self.latent_codec = nn.ModuleDict(latent_codec)

    def __getitem__(self, key: str) -> LatentCodec:
        return self.latent_codec[key]

    def forward(self, y: Tensor, side_params: Tensor) -> Dict[str, Any]:
        y_ = torch.split(y, self.groups, dim=1)
        y_out_ = [{}] * len(self.groups)
        y_hat_ = [Tensor()] * len(self.groups)
        y_likelihoods_ = [Tensor()] * len(self.groups)

        for k in range(len(self.groups)):
            params = self._get_ctx_params(k, side_params, y_hat_)
            y_out_[k] = self.latent_codec[f"y{k}"](y_[k], params)
            y_hat_[k] = y_out_[k]["y_hat"]
            y_likelihoods_[k] = y_out_[k]["likelihoods"]["y"]

        y_hat = torch.cat(y_hat_, dim=1)
        y_likelihoods = torch.cat(y_likelihoods_, dim=1)

        return {
            "likelihoods": {
                "y": y_likelihoods,
            },
            "y_hat": y_hat,
        }

    def compress(self, y: Tensor, side_params: Tensor) -> Dict[str, Any]:
        y_ = torch.split(y, self.groups, dim=1)
        y_out_ = [{}] * len(self.groups)
        y_hat = torch.zeros_like(y)
        y_hat_ = y_hat.split(self.groups, dim=1)

        for k in range(len(self.groups)):
            params = self._get_ctx_params(k, side_params, y_hat_)
            y_out_[k] = self.latent_codec[f"y{k}"].compress(y_[k], params)
            y_hat_[k][:] = y_out_[k]["y_hat"]

        y_strings_groups = [y_out["strings"] for y_out in y_out_]
        assert all(len(y_strings_groups[0]) == len(ss) for ss in y_strings_groups)

        return {
            "strings": [s for ss in y_strings_groups for s in ss],
            "shape": [y_out["shape"] for y_out in y_out_],
            "y_hat": y_hat,
        }

    def decompress(
        self,
        strings: List[List[bytes]],
        shape: List[Tuple[int, ...]],
        side_params: Tensor,
        **kwargs,
    ) -> Dict[str, Any]:
        n = len(strings[0])
        assert all(len(ss) == n for ss in strings)
        strings_per_group = len(strings) // len(self.groups)

        y_out_ = [{}] * len(self.groups)
        y_shape = (sum(s[0] for s in shape), *shape[0][1:])
        y_hat = torch.zeros((n, *y_shape), device=side_params.device)
        y_hat_ = y_hat.split(self.groups, dim=1)

        for k in range(len(self.groups)):
            params = self._get_ctx_params(k, side_params, y_hat_)
            y_out_[k] = self.latent_codec[f"y{k}"].decompress(
                strings[strings_per_group * k : strings_per_group * (k + 1)],
                shape[k],
                params,
            )
            y_hat_[k][:] = y_out_[k]["y_hat"]

        return {
            "y_hat": y_hat,
        }

    def merge_y(self, *args):
        return torch.cat(args, dim=1)

    def merge_params(self, *args):
        return torch.cat(args, dim=1)

    def _get_ctx_params(
        self, k: int, side_params: Tensor, y_hat_: List[Tensor]
    ) -> Tensor:
        if k == 0:
            return side_params
        ch_ctx_params = self.channel_context[f"y{k}"](self.merge_y(*y_hat_[:k]))
        return self.merge_params(ch_ctx_params, side_params)



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Dict, List, Mapping, Tuple

from torch import Tensor

from compressai.registry import register_module

from ..base import LatentCodec

__all__ = [
    "GainHyperpriorLatentCodec",
]


@register_module("GainHyperpriorLatentCodec")
class GainHyperpriorLatentCodec(LatentCodec):
    """Hyperprior codec constructed from latent codec for ``y`` that
    compresses ``y`` using ``params`` from ``hyper`` branch.

    Gain-controlled hyperprior introduced in
    `"Asymmetric Gained Deep Image Compression With Continuous Rate Adaptation"
    <https://arxiv.org/abs/2003.02012>`_, by Ze Cui, Jing Wang,
    Shangyin Gao, Bo Bai, Tiansheng Guo, and Yihui Feng, CVPR, 2021.

    .. code-block:: none

                z_gain  z_gain_inv
                   │        │
                   ▼        ▼
                  ┌┴────────┴┐
            ┌──►──┤ lc_hyper ├──►─┐
            │     └──────────┘    │
            │                     │
            │     y_gain          ▼ params   y_gain_inv
            │        │            │              │
            │        ▼            │              ▼
            │        │         ┌──┴───┐          │
        y ──┴────►───×───►─────┤ lc_y ├────►─────×─────►── y_hat
                               └──────┘

    By default, the following codec is constructed:

    .. code-block:: none

                        z_gain                      z_gain_inv
                           │                             │
                           ▼                             ▼
                 ┌───┐  z  │ z_g ┌───┐ z_hat      z_hat  │       ┌───┐
            ┌─►──┤h_a├──►──×──►──┤ Q ├───►───····───►────×────►──┤h_s├──┐
            │    └───┘           └───┘        EB                 └───┘  │
            │                                                           │
            │                              ┌──────────────◄─────────────┘
            │                              │            params
            │                           ┌──┴──┐
            │    y_gain                 │  EP │    y_gain_inv
            │       │                   └──┬──┘        │
            │       ▼                      │           ▼
            │       │       ┌───┐          ▼           │
        y ──┴───►───×───►───┤ Q ├────►────····───►─────×─────►── y_hat
                            └───┘          GC

    Common configurations of latent codecs include:
     - entropy bottleneck ``hyper`` (default) and gaussian conditional ``y`` (default)
     - entropy bottleneck ``hyper`` (default) and autoregressive ``y``
    """

    def __init__(self, latent_codec: Mapping[str, LatentCodec], **kwargs):
        super().__init__()
        self.y = latent_codec["y"]
        self.hyper = latent_codec["hyper"]
        self.latent_codec = latent_codec

    def __getitem__(self, key: str) -> LatentCodec:
        return self.latent_codec[key]

    def forward(
        self,
        y: Tensor,
        y_gain: Tensor,
        z_gain: Tensor,
        y_gain_inv: Tensor,
        z_gain_inv: Tensor,
    ) -> Dict[str, Any]:
        hyper_out = self.latent_codec["hyper"](y, z_gain, z_gain_inv)
        y_out = self.latent_codec["y"](y * y_gain, hyper_out["params"])
        y_hat = y_out["y_hat"] * y_gain_inv
        return {
            "likelihoods": {
                "y": y_out["likelihoods"]["y"],
                "z": hyper_out["likelihoods"]["z"],
            },
            "y_hat": y_hat,
        }

    def compress(
        self,
        y: Tensor,
        y_gain: Tensor,
        z_gain: Tensor,
        y_gain_inv: Tensor,
        z_gain_inv: Tensor,
    ) -> Dict[str, Any]:
        hyper_out = self.latent_codec["hyper"].compress(y, z_gain, z_gain_inv)
        y_out = self.latent_codec["y"].compress(y * y_gain, hyper_out["params"])
        y_hat = y_out["y_hat"] * y_gain_inv
        return {
            "strings": [*y_out["strings"], *hyper_out["strings"]],
            "shape": {"y": y_out["shape"], "hyper": hyper_out["shape"]},
            "y_hat": y_hat,
        }

    def decompress(
        self,
        strings: List[List[bytes]],
        shape: Dict[str, Tuple[int, ...]],
        y_gain_inv: Tensor,
        z_gain_inv: Tensor,
        **kwargs,
    ) -> Dict[str, Any]:
        *y_strings_, z_strings = strings
        assert all(len(y_strings) == len(z_strings) for y_strings in y_strings_)
        hyper_out = self.latent_codec["hyper"].decompress(
            [z_strings], shape["hyper"], z_gain_inv
        )
        y_out = self.latent_codec["y"].decompress(
            y_strings_, shape["y"], hyper_out["params"]
        )
        y_hat = y_out["y_hat"] * y_gain_inv
        return {"y_hat": y_hat}



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Collect performance metrics of published traditional or end-to-end image
codecs.
"""

import argparse
import json
import multiprocessing as mp
import sys

from collections import defaultdict
from itertools import starmap
from pathlib import Path
from typing import List

from .codecs import AV1, BPG, HM, JPEG, JPEG2000, TFCI, VTM, Codec, WebP

# from torchvision.datasets.folder
IMG_EXTENSIONS = (
    ".jpg",
    ".jpeg",
    ".png",
    ".ppm",
    ".bmp",
    ".pgm",
    ".tif",
    ".tiff",
    ".webp",
)

codecs = [JPEG, WebP, JPEG2000, BPG, TFCI, VTM, HM, AV1]


# we need the quality index (not value) to compute the stats later
def func(codec, i, *args):
    rv = codec.run(*args)
    return i, rv


def collect(
    codec: Codec,
    dataset: str,
    qps: List[int],
    metrics: List[str],
    num_jobs: int = 1,
):
    if not Path(dataset).is_dir():
        raise OSError(f"No such directory: {dataset}")

    filepaths = []
    for ext in IMG_EXTENSIONS:
        filepaths.extend(Path(dataset).rglob(f"*{ext}"))

    pool = mp.Pool(num_jobs) if num_jobs > 1 else None

    if len(filepaths) == 0:
        print("No images found in the dataset directory")
        sys.exit(1)

    args = [
        (codec, i, f, q, metrics) for i, q in enumerate(qps) for f in sorted(filepaths)
    ]

    if pool:
        rv = pool.starmap(func, args)
    else:
        rv = list(starmap(func, args))

    results = [defaultdict(float) for _ in range(len(qps))]

    for i, metrics in rv:
        for k, v in metrics.items():
            results[i][k] += v

    # aggregate results for all images
    for i, _ in enumerate(results):
        for k, v in results[i].items():
            results[i][k] = v / len(filepaths)

    # list of dict -> dict of list
    out = defaultdict(list)
    for r in results:
        for k, v in r.items():
            out[k].append(v)
    return out


def setup_args():
    description = "Collect codec metrics."
    parser = argparse.ArgumentParser(description=description)
    subparsers = parser.add_subparsers(dest="codec", help="Select codec")
    subparsers.required = True
    return parser, subparsers


def setup_common_args(parser):
    parser.add_argument("dataset", type=str)
    parser.add_argument(
        "-j",
        "--num-jobs",
        type=int,
        metavar="N",
        default=1,
        help="number of parallel jobs (default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--qps",
        dest="qps",
        type=str,
        default="75",
        help="list of quality/quantization parameter (default: %(default)s)",
    )
    parser.add_argument(
        "--metrics",
        dest="metrics",
        default=["psnr-rgb", "ms-ssim-rgb"],
        nargs="+",
        help="do not return PSNR and MS-SSIM metrics (use for very small images)",
    )


def main(argv):
    parser, subparsers = setup_args()
    for c in codecs:
        cparser = subparsers.add_parser(c.__name__.lower(), help=f"{c.__name__}")
        setup_common_args(cparser)
        c.setup_args(cparser)
    args = parser.parse_args(argv)

    codec_cls = next(c for c in codecs if c.__name__.lower() == args.codec)
    codec = codec_cls(args)
    qps = [int(q) for q in args.qps.split(",") if q]
    results = collect(
        codec,
        args.dataset,
        sorted(qps),
        args.metrics,
        args.num_jobs,
    )

    output = {
        "name": codec.name,
        "description": codec.description,
        "results": results,
    }

    print(json.dumps(output, indent=2))


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy

import pytest
import torch

from compressai.entropy_models import (
    EntropyBottleneckVbr,
    EntropyModelVbr,
    GaussianConditional,
)
from compressai.zoo import bmshj2018_hyperprior_vbr


@pytest.fixture
def entropy_model():
    return EntropyModelVbr()


class TestEntropyModelVbr:
    def test_quantize_invalid(self, entropy_model):
        x = torch.rand(1, 3, 4, 4)
        with pytest.raises(ValueError):
            entropy_model.quantize(x, mode="toto")

    def test_quantize_noise(self, entropy_model):
        x = torch.rand(1, 3, 4, 4)
        y = entropy_model.quantize(x, "noise")

        assert y.shape == x.shape
        assert ((y - x) <= 0.5).all()
        assert ((y - x) >= -0.5).all()
        assert (y != torch.round(x)).any()

    def test_quantize(self, entropy_model):
        x = torch.rand(1, 3, 4, 4)
        s = torch.rand(1).item()
        torch.manual_seed(s)
        y0 = entropy_model.quantize(x, "noise")
        torch.manual_seed(s)

        with pytest.warns(UserWarning):
            y1 = entropy_model._quantize(x, "noise")
        assert (y0 == y1).all()

    def test_quantize_symbols(self, entropy_model):
        x = torch.rand(1, 3, 4, 4)
        y = entropy_model.quantize(x, "symbols")

        assert y.shape == x.shape
        assert (y == torch.round(x).int()).all()

    def test_quantize_dequantize(self, entropy_model):
        x = torch.rand(1, 3, 4, 4)
        means = torch.rand(1, 3, 4, 4)
        y = entropy_model.quantize(x, "dequantize", means)

        assert y.shape == x.shape
        assert (y == torch.round(x - means) + means).all()

    def test_dequantize(self, entropy_model):
        x = torch.randint(-32, 32, (1, 3, 4, 4))
        means = torch.rand(1, 3, 4, 4)
        y = entropy_model.dequantize(x, means)

        assert y.shape == x.shape
        assert y.type() == means.type()

        with pytest.warns(UserWarning):
            yy = entropy_model._dequantize(x, means)
        assert (yy == y).all()

    def test_forward(self, entropy_model):
        with pytest.raises(NotImplementedError):
            entropy_model()

    def test_invalid_coder(self):
        with pytest.raises(ValueError):
            EntropyModelVbr(entropy_coder="huffman")

        with pytest.raises(ValueError):
            EntropyModelVbr(entropy_coder=0xFF)

    def test_invalid_inputs(self, entropy_model):
        with pytest.raises(TypeError):
            entropy_model.compress(torch.rand(1, 3))
        with pytest.raises(ValueError):
            entropy_model.compress(torch.rand(1, 3), torch.rand(2, 3))
        with pytest.raises(ValueError):
            entropy_model.compress(torch.rand(1, 3, 1, 1), torch.rand(2, 3))

    def test_invalid_cdf(self, entropy_model):
        x = torch.rand(1, 32, 16, 16)
        indexes = torch.rand(1, 32, 16, 16)
        with pytest.raises(ValueError):
            entropy_model.compress(x, indexes)

    def test_invalid_cdf_length(self, entropy_model):
        x = torch.rand(1, 32, 16, 16)
        indexes = torch.rand(1, 32, 16, 16)
        entropy_model._quantized_cdf.resize_(32, 1)

        with pytest.raises(ValueError):
            entropy_model.compress(x, indexes)

        entropy_model._cdf_length.resize_(32, 1)
        with pytest.raises(ValueError):
            entropy_model.compress(x, indexes)

    def test_invalid_offsets(self, entropy_model):
        x = torch.rand(1, 32, 16, 16)
        indexes = torch.rand(1, 32, 16, 16)
        entropy_model._quantized_cdf.resize_(32, 1)
        entropy_model._cdf_length.resize_(32)
        with pytest.raises(ValueError):
            entropy_model.compress(x, indexes)

    def test_invalid_decompress(self, entropy_model):
        with pytest.raises(TypeError):
            entropy_model.decompress(["ssss"])

        with pytest.raises(ValueError):
            entropy_model.decompress("sss", torch.rand(1, 3, 4, 4))

        with pytest.raises(ValueError):
            entropy_model.decompress(["sss"], torch.rand(1, 4, 4))

        with pytest.raises(ValueError):
            entropy_model.decompress(["sss"], torch.rand(2, 4, 4))

        with pytest.raises(ValueError):
            entropy_model.decompress(["sss"], torch.rand(1, 4, 4), torch.rand(2, 4, 4))


class TestEntropyBottleneckVbr:
    def test_forward_training(self):
        entropy_bottleneck = EntropyBottleneckVbr(128)
        x = torch.rand(1, 128, 32, 32)
        y, y_likelihoods = entropy_bottleneck(x)

        assert isinstance(entropy_bottleneck, EntropyModelVbr)
        assert y.shape == x.shape
        assert y_likelihoods.shape == x.shape

        assert ((y - x) <= 0.5).all()
        assert ((y - x) >= -0.5).all()
        assert (y != torch.round(x)).any()

    def test_forward_inference_0D(self):
        entropy_bottleneck = EntropyBottleneckVbr(128)
        entropy_bottleneck.eval()
        x = torch.rand(1, 128)
        y, y_likelihoods = entropy_bottleneck(x)

        assert y.shape == x.shape
        assert y_likelihoods.shape == x.shape

        assert (y == torch.round(x)).all()

    def test_forward_inference_2D(self):
        entropy_bottleneck = EntropyBottleneckVbr(128)
        entropy_bottleneck.eval()
        x = torch.rand(1, 128, 32, 32)
        y, y_likelihoods = entropy_bottleneck(x)

        assert y.shape == x.shape
        assert y_likelihoods.shape == x.shape

        assert (y == torch.round(x)).all()

    def test_forward_inference_ND(self):
        entropy_bottleneck = EntropyBottleneckVbr(128)
        entropy_bottleneck.eval()

        # Test 0D
        x = torch.rand(1, 128)
        y, y_likelihoods = entropy_bottleneck(x)

        assert y.shape == x.shape
        assert y_likelihoods.shape == x.shape

        assert (y == torch.round(x)).all()

        # Test from 1 to 5 dimensions
        for i in range(1, 6):
            x = torch.rand(1, 128, *([4] * i))
            y, y_likelihoods = entropy_bottleneck(x)

            assert y.shape == x.shape
            assert y_likelihoods.shape == x.shape

            assert (y == torch.round(x)).all()

    def test_loss(self):
        entropy_bottleneck = EntropyBottleneckVbr(128)
        loss = entropy_bottleneck.loss()

        assert len(loss.size()) == 0
        assert loss.numel() == 1

    def test_update(self):
        # get a pretrained model
        net = bmshj2018_hyperprior_vbr(quality=0, pretrained=True).eval()
        # update works in force mode only
        # assert not net.update(scale=1)
        # assert not net.update(force=False)
        assert net.update(force=True)

    def test_compression_2D(self):
        x = torch.rand(1, 128, 32, 32)
        eb = EntropyBottleneckVbr(128)
        eb.update()
        s = eb.compress(x)
        x2 = eb.decompress(s, x.size()[2:])
        means = eb._get_medians()

        assert torch.allclose(torch.round(x - means) + means, x2)

    def test_compression_ND(self):
        eb = EntropyBottleneckVbr(128)
        eb.update()

        # Test 0D
        x = torch.rand(1, 128)
        s = eb.compress(x)
        x2 = eb.decompress(s, [])
        means = eb._get_medians().reshape(128)

        assert torch.allclose(torch.round(x - means) + means, x2)

        # Test from 1 to 5 dimensions
        for i in range(1, 6):
            x = torch.rand(1, 128, *([4] * i))
            s = eb.compress(x)
            x2 = eb.decompress(s, x.size()[2:])
            means = eb._get_medians().reshape(128, *([1] * i))

            assert torch.allclose(torch.round(x - means) + means, x2)


@pytest.mark.parametrize(
    "model_cls,args",
    ((EntropyBottleneckVbr, (128,)),),
)
def test_deepcopy(model_cls, args):
    model = model_cls(*args)
    model_copy = copy.deepcopy(model)
    x = torch.rand(1, 128, 32, 32)

    if isinstance(model, GaussianConditional):
        opts = (torch.rand_like(x), torch.rand_like(x))
    else:
        opts = ()

    torch.manual_seed(32)
    y0 = model(x, *opts)

    torch.manual_seed(32)
    y1 = model_copy(x, *opts)

    assert torch.allclose(y0[0], y1[0])



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import torch.nn as nn

from compressai.layers.basic import Gain, Interleave, Reshape

GAIN = 10.0


def conv1d_group_seq(
    num_channels,
    groups=None,
    kernel_size=1,
    enabled=("bn", "act"),
    enabled_final=("bn", "act"),
):
    if groups is None:
        groups = [1] * (len(num_channels) - 1)
    assert len(num_channels) == 0 or len(groups) == len(num_channels) - 1
    xs = []
    for i in range(len(num_channels) - 1):
        is_final = i + 1 == len(num_channels) - 1
        xs.append(
            nn.Conv1d(
                num_channels[i], num_channels[i + 1], kernel_size, groups=groups[i]
            )
        )
        # ChannelShuffle is only required between consecutive group convs.
        if not is_final and groups[i] > 1 and groups[i + 1] > 1:
            xs.append(Interleave(groups[i]))
        if "bn" in enabled and (not is_final or "bn" in enabled_final):
            xs.append(nn.BatchNorm1d(num_channels[i + 1]))
        if "act" in enabled and (not is_final or "act" in enabled_final):
            xs.append(nn.ReLU(inplace=True))
    return nn.Sequential(*xs)


def pointnet_g_a_simple(num_channels, groups=None, gain=GAIN):
    return nn.Sequential(
        *conv1d_group_seq(num_channels, groups),
        nn.AdaptiveMaxPool1d(1),
        Gain((num_channels[-1], 1), gain),
    )


def pointnet_g_s_simple(num_channels, gain=GAIN):
    return nn.Sequential(
        Gain((num_channels[0], 1), 1 / gain),
        *conv1d_group_seq(num_channels, enabled=["act"], enabled_final=[]),
        Reshape((num_channels[-1] // 3, 3)),
    )


def pointnet_classification_backend(num_channels):
    return nn.Sequential(
        *conv1d_group_seq(num_channels[:-1], enabled_final=[]),
        nn.Dropout(0.3),
        nn.Conv1d(num_channels[-2], num_channels[-1], 1),
        Reshape((num_channels[-1],)),
    )



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import torch
import torch.nn as nn

from torch import Tensor

__all__ = [
    "Lambda",
    "NamedLayer",
    "Reshape",
    "Transpose",
    "Interleave",
    "Gain",
]


class Lambda(nn.Module):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __repr__(self):
        return f"{self.__class__.__name__}(func={self.func})"

    def forward(self, x):
        return self.func(x)


class NamedLayer(nn.Module):
    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"

    def forward(self, x):
        return x


class Reshape(nn.Module):
    def __init__(self, shape):
        super().__init__()
        self.shape = shape

    def __repr__(self):
        return f"{self.__class__.__name__}(shape={self.shape})"

    def forward(self, x):
        output_shape = (x.shape[0], *self.shape)
        try:
            return x.reshape(output_shape)
        except RuntimeError as e:
            e.args += (f"Cannot reshape input {tuple(x.shape)} to {output_shape}",)
            raise e


class Transpose(nn.Module):
    def __init__(self, dim0, dim1):
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1

    def __repr__(self):
        return f"{self.__class__.__name__}(dim0={self.dim0}, dim1={self.dim1})"

    def forward(self, x):
        return x.transpose(self.dim0, self.dim1).contiguous()


class Interleave(nn.Module):
    def __init__(self, groups: int):
        super().__init__()
        self.groups = groups

    def forward(self, x: Tensor) -> Tensor:
        g = self.groups
        n, c, *tail = x.shape
        return x.reshape(n, g, c // g, *tail).transpose(1, 2).reshape(x.shape)


class Gain(nn.Module):
    def __init__(self, shape=None, factor: float = 1.0):
        super().__init__()
        self.factor = factor
        self.gain = nn.Parameter(torch.ones(shape))

    def forward(self, x: Tensor) -> Tensor:
        return self.factor * self.gain * x



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from compressai.latent_codecs import LatentCodec
from compressai.latent_codecs.entropy_bottleneck import EntropyBottleneckLatentCodec
from compressai.layers.pointcloud.pointnet import (
    pointnet_g_a_simple,
    pointnet_g_s_simple,
)
from compressai.models import CompressionModel
from compressai.registry import register_model

__all__ = [
    "PointNetReconstructionPccModel",
]


@register_model("sfu2023-pcc-rec-pointnet")
class PointNetReconstructionPccModel(CompressionModel):
    """PointNet-based PCC reconstruction model.

    Model based on PointNet [Qi2017PointNet]_, modified for compression
    by [Yan2019]_, with layer configurations and other modifications as
    used in [Ulhaq2023]_.

    References:

        .. [Qi2017PointNet] `"PointNet: Deep Learning on Point Sets for
            3D Classification and Segmentation"
            <https://arxiv.org/abs/1612.00593>`_, by Charles R. Qi,
            Hao Su, Kaichun Mo, and Leonidas J. Guibas, CVPR 2017.

        .. [Yan2019] `"Deep AutoEncoder-based Lossy Geometry Compression
            for Point Clouds" <https://arxiv.org/abs/1905.03691>`_,
            by Wei Yan, Yiting Shao, Shan Liu, Thomas H Li, Zhu Li,
            and Ge Li, 2019.

        .. [Ulhaq2023] `"Learned Point Cloud Compression for
            Classification" <https://arxiv.org/abs/2308.05959>`_,
            by Mateen Ulhaq and Ivan V. Bajić, MMSP 2023.
    """

    latent_codec: LatentCodec

    def __init__(
        self,
        num_points=1024,
        num_channels={  # noqa: B006
            "g_a": [3, 64, 64, 64, 128, 1024],
            "g_s": [1024, 256, 512, 1024 * 3],
        },
        groups={  # noqa: B006
            "g_a": [1, 1, 1, 1, 1],
        },
    ):
        super().__init__()

        assert num_channels["g_a"][-1] == num_channels["g_s"][0]
        assert num_channels["g_s"][-1] == num_points * 3

        self.g_a = pointnet_g_a_simple(num_channels["g_a"], groups["g_a"])

        self.g_s = pointnet_g_s_simple(num_channels["g_s"])

        self.latent_codec = EntropyBottleneckLatentCodec(
            channels=num_channels["g_a"][-1],
            tail_mass=1e-4,
        )

    def forward(self, input):
        x = input["pos"]
        x_t = x.transpose(-2, -1)
        y = self.g_a(x_t)
        y_out = self.latent_codec(y)
        y_hat = y_out["y_hat"]
        x_hat = self.g_s(y_hat)
        assert x_hat.shape == x.shape

        return {
            "x_hat": x_hat,
            "likelihoods": {
                "y": y_out["likelihoods"]["y"],
            },
            # Additional outputs:
            "y": y,
            "y_hat": y_hat,
            "debug_outputs": {
                "y_hat": y_hat,
            },
        }

    def compress(self, input):
        x = input["pos"]
        x_t = x.transpose(-2, -1)
        y = self.g_a(x_t)
        y_out = self.latent_codec.compress(y)
        [y_strings] = y_out["strings"]
        return {"strings": [y_strings], "shape": (1,)}

    def decompress(self, strings, shape):
        assert isinstance(strings, list) and len(strings) == 1
        [y_strings] = strings
        y_hat = self.latent_codec.decompress([y_strings], shape)
        x_hat = self.g_s(y_hat)
        return {"x_hat": x_hat}



# Copyright (c) 2021-2024, InterDigital Communications, Inc
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the disclaimer
# below) provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of InterDigital Communications, Inc nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Callable, Dict, List, Tuple, TypeVar

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from compressai.ans import BufferedRansEncoder, RansDecoder
from compressai.entropy_models import GaussianConditional
from compressai.layers import MaskedConv2d
from compressai.registry import register_module

from .base import LatentCodec

__all__ = [
    "RasterScanLatentCodec",
]

K = TypeVar("K")
V = TypeVar("V")


@register_module("RasterScanLatentCodec")
class RasterScanLatentCodec(LatentCodec):
    """Autoregression in raster-scan order with local decoded context.

    PixelCNN context model introduced in
    `"Pixel Recurrent Neural Networks"
    <http://arxiv.org/abs/1601.06759>`_,
    by Aaron van den Oord, Nal Kalchbrenner, and Koray Kavukcuoglu,
    International Conference on Machine Learning (ICML), 2016.

    First applied to learned image compression in
    `"Joint Autoregressive and Hierarchical Priors for Learned Image
    Compression" <https://arxiv.org/abs/1809.02736>`_,
    by D. Minnen, J. Balle, and G.D. Toderici,
    Adv. in Neural Information Processing Systems 31 (NeurIPS 2018).

    .. code-block:: none

                         ctx_params
                             │
                             ▼
                             │ ┌───◄───┐
                           ┌─┴─┴─┐  ┌──┴──┐
                           │  EP │  │  CP │
                           └──┬──┘  └──┬──┘
                              │        │
                              │        ▲
               ┌───┐  y_hat   ▼        │
        y ──►──┤ Q ├────►────····───►──┴──►── y_hat
               └───┘          GC

    """

    def __init__(
        self,
        gaussian_conditional: GaussianConditional,
        entropy_parameters: nn.Module,
        context_prediction: MaskedConv2d,
        **kwargs,
    ):
        super().__init__()
        self.gaussian_conditional = gaussian_conditional
        self.entropy_parameters = entropy_parameters
        self.context_prediction = context_prediction
        self.kernel_size = _to_single(self.context_prediction.kernel_size)
        self.padding = (self.kernel_size - 1) // 2

    def forward(self, y: Tensor, params: Tensor) -> Dict[str, Any]:
        y_hat = self.gaussian_conditional.quantize(
            y, "noise" if self.training else "dequantize"
        )
        ctx_params = self.merge(params, self.context_prediction(y_hat))
        gaussian_params = self.entropy_parameters(ctx_params)
        scales_hat, means_hat = gaussian_params.chunk(2, 1)
        _, y_likelihoods = self.gaussian_conditional(y, scales_hat, means=means_hat)
        return {"likelihoods": {"y": y_likelihoods}, "y_hat": y_hat}

    def compress(self, y: Tensor, ctx_params: Tensor) -> Dict[str, Any]:
        n, _, y_height, y_width = y.shape
        ds = []
        for i in range(n):
            encoder = BufferedRansEncoder()
            y_hat = raster_scan_compress_single_stream(
                encoder=encoder,
                y=y[i : i + 1, :, :, :],
                params=ctx_params[i : i + 1, :, :, :],
                gaussian_conditional=self.gaussian_conditional,
                entropy_parameters=self.entropy_parameters,
                context_prediction=self.context_prediction,
                height=y_height,
                width=y_width,
                padding=self.padding,
                kernel_size=self.kernel_size,
                merge=self.merge,
            )
            y_strings = encoder.flush()
            ds.append({"strings": [y_strings], "y_hat": y_hat.squeeze(0)})
        return {**default_collate(ds), "shape": y.shape[2:4]}

    def decompress(
        self,
        strings: List[List[bytes]],
        shape: Tuple[int, int],
        ctx_params: Tensor,
        **kwargs,
    ) -> Dict[str, Any]:
        (y_strings,) = strings
        y_height, y_width = shape
        ds = []
        for i in range(len(y_strings)):
            decoder = RansDecoder()
            decoder.set_stream(y_strings[i])
            y_hat = raster_scan_decompress_single_stream(
                decoder=decoder,
                params=ctx_params[i : i + 1, :, :, :],
                gaussian_conditional=self.gaussian_conditional,
                entropy_parameters=self.entropy_parameters,
                context_prediction=self.context_prediction,
                height=y_height,
                width=y_width,
                padding=self.padding,
                kernel_size=self.kernel_size,
                device=ctx_params.device,
                merge=self.merge,
            )
            ds.append({"y_hat": y_hat.squeeze(0)})
        return default_collate(ds)

    @staticmethod
    def merge(*args):
        return torch.cat(args, dim=1)


def raster_scan_compress_single_stream(
    encoder: BufferedRansEncoder,
    y: Tensor,
    params: Tensor,
    *,
    gaussian_conditional: GaussianConditional,
    entropy_parameters: nn.Module,
    context_prediction: MaskedConv2d,
    height: int,
    width: int,
    padding: int,
    kernel_size: int,
    merge: Callable[..., Tensor] = lambda *args: torch.cat(args, dim=1),
) -> Tensor:
    """Compresses y and writes to encoder bitstream.

    Returns:
        The y_hat that will be reconstructed at the decoder.
    """
    assert height == y.shape[-2]
    assert width == y.shape[-1]

    cdf = gaussian_conditional.quantized_cdf.tolist()
    cdf_lengths = gaussian_conditional.cdf_length.tolist()
    offsets = gaussian_conditional.offset.tolist()
    masked_weight = context_prediction.weight * context_prediction.mask

    y_hat = _pad_2d(y, padding)

    symbols_list = []
    indexes_list = []

    # Warning, this is slow...
    # TODO: profile the calls to the bindings...
    for h in range(height):
        for w in range(width):
            # only perform the mask convolution on a cropped tensor
            # centered in (h, w)
            y_crop = y_hat[:, :, h : h + kernel_size, w : w + kernel_size]
            ctx_p = F.conv2d(
                y_crop,
                masked_weight,
                context_prediction.bias,
            )

            # 1x1 conv for the entropy parameters prediction network, so
            # we only keep the elements in the "center"
            p = params[:, :, h : h + 1, w : w + 1]
            gaussian_params = entropy_parameters(merge(p, ctx_p))
            gaussian_params = gaussian_params.squeeze(3).squeeze(2)
            scales_hat, means_hat = gaussian_params.chunk(2, 1)
            indexes = gaussian_conditional.build_indexes(scales_hat)

            y_crop = y_crop[:, :, padding, padding]
            symbols = gaussian_conditional.quantize(y_crop, "symbols", means_hat)
            y_hat_item = symbols + means_hat

            hp = h + padding
            wp = w + padding
            y_hat[:, :, hp, wp] = y_hat_item

            symbols_list.extend(symbols.squeeze().tolist())
            indexes_list.extend(indexes.squeeze().tolist())

    encoder.encode_with_indexes(symbols_list, indexes_list, cdf, cdf_lengths, offsets)

    y_hat = _pad_2d(y_hat, -padding)
    return y_hat


def raster_scan_decompress_single_stream(
    decoder: RansDecoder,
    params: Tensor,
    *,
    gaussian_conditional: GaussianConditional,
    entropy_parameters: nn.Module,
    context_prediction: MaskedConv2d,
    height: int,
    width: int,
    padding: int,
    kernel_size: int,
    device,
    merge: Callable[..., Tensor] = lambda *args: torch.cat(args, dim=1),
) -> Tensor:
    """Decodes y_hat from decoder bitstream.

    Returns:
        The reconstructed y_hat.
    """
    cdf = gaussian_conditional.quantized_cdf.tolist()
    cdf_lengths = gaussian_conditional.cdf_length.tolist()
    offsets = gaussian_conditional.offset.tolist()
    masked_weight = context_prediction.weight * context_prediction.mask

    c = context_prediction.in_channels
    shape = (1, c, height + 2 * padding, width + 2 * padding)
    y_hat = torch.zeros(shape, device=device)

    # Warning: this is slow due to the auto-regressive nature of the
    # decoding... See more recent publication where they use an
    # auto-regressive module on chunks of channels for faster decoding...
    for h in range(height):
        for w in range(width):
            # only perform the mask convolution on a cropped tensor
            # centered in (h, w)
            y_crop = y_hat[:, :, h : h + kernel_size, w : w + kernel_size]
            ctx_p = F.conv2d(
                y_crop,
                masked_weight,
                context_prediction.bias,
            )

            # 1x1 conv for the entropy parameters prediction network, so
            # we only keep the elements in the "center"
            p = params[:, :, h : h + 1, w : w + 1]
            gaussian_params = entropy_parameters(merge(p, ctx_p))
            gaussian_params = gaussian_params.squeeze(3).squeeze(2)
            scales_hat, means_hat = gaussian_params.chunk(2, 1)
            indexes = gaussian_conditional.build_indexes(scales_hat)

            symbols = decoder.decode_stream(
                indexes.squeeze().tolist(), cdf, cdf_lengths, offsets
            )
            symbols = Tensor(symbols).reshape(1, -1)
            y_hat_item = gaussian_conditional.dequantize(symbols, means_hat)

            hp = h + padding
            wp = w + padding
            y_hat[:, :, hp, wp] = y_hat_item

    y_hat = _pad_2d(y_hat, -padding)
    return y_hat


def _pad_2d(x: Tensor, padding: int) -> Tensor:
    return F.pad(x, (padding, padding, padding, padding))


def _to_single(xs):
    assert all(x == xs[0] for x in xs)
    return xs[0]


def default_collate(batch: List[Dict[K, V]]) -> Dict[K, List[V]]:
    """Combines a list of dictionaries into a single dictionary.

    Workaround to ``torch.utils.data.default_collate`` bug in PyTorch 2.0.0.
    """
    if not isinstance(batch, list) or any(not isinstance(d, dict) for d in batch):
        raise NotImplementedError

    result = _ld_to_dl(batch)

    for k, vs in result.items():
        if all(isinstance(v, Tensor) for v in vs):
            result[k] = torch.stack(vs)

    return result


def _ld_to_dl(ld: List[Dict[K, V]]) -> Dict[K, List[V]]:
    dl = {}
    for d in ld:
        for k, v in d.items():
            if k not in dl:
                dl[k] = []
            dl[k].append(v)
    return dl
